//
// Grid.java
//

package ps5;

public class Grid {
	// OVERVIEW: A Grid is a 2-dimensional grid.  Each square in the grid is either
	//    empty or contains a SimObject object.
	//
	//    A typical Grid is: [ [ s_0_ymax - 1, s_1_ymax - 1, ..., s_xmax - 1_ymax - 1 ],
	//						   ... ,
	//                         [ s_0_1, s_1_1, ..., s_xmax - 1_1],
	//                         [ s_0_0, s_1_0, ..., s_xmax - 1_0] ]
	//						 
	//    Where each s_x_y is either a SimObject object or is empty.
	//

	protected int xsize;
	protected int ysize;
	SimObject[][] simobjects;
	
	// Rep Invariant 
	//     simobjects.length == xsize;
	//     forall 0 <= i < xsize, simobjects[i].length == ysize;
	//     simobjects[x][y] == null      
	//     or simobjects[x][y].getX() == x && simobjects[x][y].getY() == y

	Grid(int p_xsize, int p_ysize)
	// REQUIRES p_xsize > 0 and p_ysize > 0
	// EFFECTS: Constructs a new grid with horizontal size p_xsize and 
	//       vertical size p_ysize and all cells containing null.
	{
		xsize = p_xsize;
		ysize = p_ysize;
		simobjects = new SimObject[xsize][ysize];
	}

	public int numHorizontal()
	// EFFECTS: Returns the number of horizontal squares in this.
	{
		return xsize;
	}

	public int numVertical()
	// EFFECTS: Returns the number of vertical squares in this.
	{
		return ysize;
	}

	synchronized public void setObjectAt(int x, int y, SimObject obj) 
	   throws BadLocationException
    // EFFECTS: If x is outside the gird (x < 0 or x >= horizontal size, or
	//     y is outside the grid (y < 0 or y >= vertical size) throws 
	//     BadLocationException.  Otherwise, sets the object at grid
	//     location (x, y) to newObject.
	{ 
		checkValidLocation(x, y);
		simobjects[x][y] = obj; 
	} 

	public Coordinate randomFreeLocation()
	// REQUIRES: There is at least one free square on the grid
	//       The client should lock the grid before calling this, and
	//       hold the lock until the square is used, otherwise the square
	//       may be occupied before the object is placed in it.
	// EFFECTS: returns a new location on the grid that is unoccupied
	{
		// Dangerous implementation...if there is no free square, this will 
		// loop forever... (but requires was not satisfied)
		while (true) {
			int xloc = GenRandom.nextInt(xsize);
			int yloc = GenRandom.nextInt(ysize);
			
			try {
				if (getObjectAt(xloc, yloc) == null) {
					return new Coordinate(xloc, yloc);
				}
			} catch (BadLocationException ble) {
				throw new RuntimeException (ble); // should never happen
			}
		}
	} 

	synchronized public void setObjectAt(Coordinate coord, SimObject obj) 
	    throws BadLocationException
	// REQUIRES: coord is not null and obj is not null    
	// EFFECTS: If coord is a location on grid, places newObject on the grid 
	//     at the location given by coord.  
	//     Otherwise, throws BadLocationException.
	{ 
		setObjectAt(coord.getX(), coord.getY(), obj);
	}

	synchronized public void pauseObjects()
	// EFFECTS: Pause all object threads.
	{ 
		for (int x = 0; x < numHorizontal(); x++) {
			for (int y = 0; y < numVertical(); y++) {
				SimObject current = simobjects[x][y];
				if (current != null) {
					current.pauseObject();
				}
			}
		}
	}

	synchronized public void startObjects()
	// EFFECTS: Start all object threads.  Previously started
	//	threads will be resumed.
	{ 
		for (int x = 0; x < numHorizontal(); x++) {
			for (int y = 0; y < numVertical(); y++) {
				SimObject current = simobjects[x][y];
				if (current != null) {
					current.resumeObject();
				}
			}
		}
	}
	
	synchronized public void removeAllObjects()
	// EFFECTS: removes all objects from the grid.
	{ 
		for (int x = 0; x < numHorizontal(); x++) {
			for (int y = 0; y < numVertical(); y++) {
				simobjects[x][y] = null;
			}
		}
	}

	synchronized public void removeObjectAt(int x, int y) throws BadLocationException
	// EFFECTS: If (x, y) is not a valid coordinate on this, throws BadLocationException.
	//    Otherwise, sets the object at (x, y) to null.
	{ 
		checkValidLocation(x, y);
		simobjects[x][y] = null;
	} 

	synchronized public void removeObjectAt(Coordinate c) throws BadLocationException
	// REQUIRES c != null
	// EFFECTS: If c is not a valid coordinate on this, throws BadLocationException.
	//    Otherwise, sets the object at (x, y) to null.
	{ 
		removeObjectAt(c.getX(), c.getY());
	}

	synchronized public boolean isSquareEmpty(int x, int y)
		throws RuntimeException, BadLocationException
	// EFFECTS: If x, y is not a valid location, throws BadLocationException.
	//     Otherwise, returns true iff there is no object at location x, y.         
	{ 
		return (getObjectAt(x, y) == null);
	}


	public boolean validLocation(Coordinate c) 
	// REQUIRES: c != null
	// ENSURES: returns true iff c is a valid location on this.
	{
		return validLocation(c.getX(), c.getY());
	}

	public boolean validLocation(int x, int y)
    // EFFECTS: Returns true iff x, y is a valid grid location.
	{
		return (x >= 0 && y >= 0 && y < numVertical () && x < numHorizontal ());
	}

	private void checkValidLocation(int x, int y) throws BadLocationException
	// EFFECTS: If <x, y> is not a valid location on this, throws BadLocationException.
	//     Otherwise, does nothing.
	{
		if (!validLocation (x, y)) {
			throw new BadLocationException ("checkValidLocation fails: " + x + ", " + y);
		}		
	}
	
	/* unsynchronized */
	public SimObject grabObjectAt(int x, int y) 
		throws RuntimeException, BadLocationException
	// We need the unsynchronized version of getObjectAt for the display --- we want to 
	// be able to display the grid without acquiring a lock.  This runs the risk that
	// the grid will be temporarily displayed in an inconsistent state.
	// EFFECTS: If <x, y> is not a valid location on this, throws BadLocationException.
	//    Otherwise, returns the (possibly null) object at <x, y>.  
	{
		checkValidLocation (x, y);
		return simobjects[x][y];
	}

	synchronized public SimObject getObjectAt(int x, int y)
		throws RuntimeException, BadLocationException
	// EFFECTS: If <x, y> is not a valid location on this, throws BadLocationException.
    //    Otherwise, returns the (possibly null) object at <x, y>.
	//    Note: identical to grabObjectAt(int,int), except this method is synchronized.
	{ 
		return grabObjectAt(x, y);
	}

	synchronized public SimObject getObjectAt(Coordinate c) 
		throws RuntimeException, BadLocationException
	// EFFECTS: If c is not a valid location on this, throws BadLocationException.
    //    Otherwise, returns the (possibly null) object at c.
	{ 
		return grabObjectAt(c.getX(), c.getY());
	}
}
