//
// SimObject.java
//

package ps5;

import java.util.Random;
import java.util.Enumeration;
import java.util.Vector;
import java.awt.Color;

abstract public class SimObject implements Runnable {
	// OVERVIEW: A SimObject is an object than represents a simulator object.
	// It has a grid (a circular reference to the grid containing this object),
	// location (integer row and column), initialization state (boolean
	// that is true if it has beein initialized), and thread pause state
	// (boolean that
	// is true if the thread is paused).
	// A typical SimObject is < grid, row, col, initialized, paused >.
	//
	// Note: SimObject is an abstract class. The executeTurn method must
	// be implemented by subclasses.

	// Constants (in milliseconds)
	static final int TURN_DELAY = 500;
	static final int TURN_RANDOM = 200;

	// Rep:
	protected Grid grid;
	private boolean isPaused;

	// subclasses have access to mrow and mcol
	protected int mx;
	protected int my;
	private Random random; // Used to get random numbers.

	// AF(c) = < grid, mrow, mcol, isInitialized, isPaused >
	// RI(c) =
	// if isIntialized then
	// grid != null
	// && grid.validLocation (mx, my)
	// && grid.getObjectAt (mx, my) == this

	public SimObject()
	// EFFECTS: Creates a new, uninitalized SimObject.
	{
		random = new Random();
	}

	final public void delay(int ms)
	// EFFECTS: Stalls for ms milliseconds.
	{
		try {
			Thread.sleep(ms);
		} catch (InterruptedException ie) {
			;
		}
	}

	abstract public void executeTurn();
	// REQUIRES: this is initialized
	// MODIFIES: this
	// EFFECTS: Executes one turn for this object.

	public Color getColor()
	// EFFECTS: Returns the color that represents this SimObject, for
	// painting the grid.
	{
		return Color.green;
	}

	final public Grid getGrid()
	// REQUIRES: this is initialized.
	// EFFECTS: Returns the grid containing this cell.
	// NOTE: The rep is exposed. The Grid associated with a SimObject is part of
	// the data abstraction.
	{
		return grid;
	}

	synchronized public Enumeration<SimObject> getNeighbors()
	// REQUIRES: this is initialized.
	// The grid must be locked as long as the result of getNeighbors is used.
	// Otherwise, the neighbors could change if another object moves!
	// EFFECTS: Returns an Enumeration that yields all the objects adjacent to
	// this
	// in the grid.
	{
		Vector<SimObject> objects = new Vector<SimObject>();

		for (int x = mx - 1; x <= mx + 1; x++) {
			for (int y = my - 1; y <= my + 1; y++) {
				if (x != mx || y != my) { // don't count yourself
					if (grid.validLocation(x, y)) {
						try {
							SimObject obj = grid.getObjectAt(x, y);

							if (obj != null) {
								objects.addElement(obj);
							}
						} catch (BadLocationException ble) {
							throw new RuntimeException (ble);
						}
					}
				}
			}
		}

		return objects.elements();
	} 

	final public int getX()
	// REQUIRES: this is initialized.
	// EFFECTS: Returns the x location that this cell is located in.
	{
		return mx;
	}

	final public int getY()
	// REQUIRES: init has previously been called for this.
	// EFFECTS: Returns the y location that this cell is located in.
	{
		return my;
	}

	final public void init(int x, int y, Grid grid) throws BadLocationException
	// REQUIRES: this is uninitialized 
	// MODIFIES: this
	// EFFECTS: If x, y is not a valid location on grid, throws BadLocationException.
	//     Otherwise, initializes the cell at row, col on grid with isPaused = true.
	// NOTE: This method is final, that means subtypes cannot override this
	//    method.
	{
		this.mx = x;
		this.my = y;
		this.grid = grid;
		this.isPaused = true;
		Thread thread = new Thread(this);
		thread.setPriority(Thread.MIN_PRIORITY);
		thread.start();
	}

	public boolean isPaused()
	// EFFECTS: Returns the value of paused.
	{
		return isPaused;
	}

	public void pauseObject()
	// EFFECTS: Sets paused to true.
	{
		isPaused = true;
	}

	public void resumeObject()
	// EFFECTS: Sets paused to false.
	{
		isPaused = false;
	}

	public void run()
	// REQUIRES: this has been initialized
    // MODIFIES: this
    // EFFECTS: Implements the thread for this object.
    //    Note: This method can do anything.
    //       For the base class, if the object is not paused,
    //       it executes one turn by calling the executeTurn method,
    //       and sleeps for a time and repeats. If the object is
    //       paused, does nothing (until the object is unpaused).
	{
		while (true) {
			if (!isPaused) {
				executeTurn();
			}
			delay(TURN_DELAY + random.nextInt(TURN_RANDOM));
		}
	}
}
